import subprocess
import os


class Indexer:
    def __init__(self, args, output_path, cat_file_path):
        self.args = args
        self.output_path = output_path
        self.cat_file_path = cat_file_path

    def index(self):
        if self.args.rg:
            threads_addon = ""
            if self.args.threads:
                threads_addon = "--threads {}".format(self.args.threads)
        
            quiet_addon = ""
            if self.args.verbosity:
                print("Running the following command:")
                print("bowtie2-build {0} {1}\n".format(self.cat_file_path, self.output_path, threads_addon))
            else:
                quiet_addon = "--quiet"

            subprocess.run("bowtie2-build {3} {2} {0} {1}".format(self.cat_file_path, self.output_path, 
                                                                  threads_addon, quiet_addon),
                           shell=True)
        else:
            raise Exception("reference genome dir not specified")



"""
 Copyright 2020, Conner Kojima <cykojima@usc.edu>

 This file is the main driver of rrap, a bionformatics pipeline that
 runs read recruitment and rpkm visualization given reference 
 genome and metagenome directories.

"""

import os
import argparse
import subprocess
from pathlib import Path
from rrap import concatenator
from rrap import indexer
from rrap import read_recruiter
from rrap import visualizer


def main():
    c = Controller()
    c.run()


class Controller:
    def __init__(self):
        self.p = argparse.ArgumentParser(prog="RRAP",
                                         description="Run read recruitment on a set of cleaned fna files")

        # argument groups
        self.inputs = None
        self.outputs = None
        self.optional = None
        self.subcommands = None
        self.arg_groups = []

        # args
        self.args = None

        # pipes
        self.concatenator = None
        self.indexer = None
        self.read_recruiter = None
        self.visualizer = None

        # intermediate products
        self.cat_file_path = None
        self.index_dir_path = None

        self.rpkm_heater_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "rpkm_heater.py")
        self.stats_dir_path = None

    def run(self):
        self.add_arguments()

        ("\n---------making output dir if needed-------------")
        self.set_output_dir()
        self.cat_file_name = os.path.join(self.args.o, 'allgenomes_cat_{0}.fna'.format(self.args.n))

        if not self.args.index_pass:
            if self.args.crg:
                self.cat_file_path = self.args.crg
            else:
                print("\n---------concatenating reference genomes-------------")
                self.concatenator = concatenator.Concatenator(self.args)
                self.cat_file_path = self.concatenator.concatenate()

            print("\n---------indexing reference genomes-------------")
            self.indexer = indexer.Indexer(self.args, os.path.join(self.index_dir_path, self.args.n), self.cat_file_path)
            self.indexer.index()
        else:
            print("\n---------skipped indexing reference genomes-------------")

        if not self.args.rr_pass:
            print("\n---------read recruitment and data transform-------------")
            self.read_recruiter = read_recruiter.ReadRecruiter(self.args, os.path.join(self.index_dir_path, self.args.n),
                                                               self.cat_file_path, self.stats_dir_path,
                                                               self.bam_dir_path)
            self.read_recruiter.read_recruit()
        else:
            print("\n---------skipped read recruitment and data transform-------------")

        if not self.args.vis_pass:
            print("\n---------visualization-------------")
            self.visualizer = visualizer.Visualizer(self.args, self.rpkm_heater_path, self.stats_dir_path)
            self.visualizer.calculate_rpkm()
            if self.args.extra_vis:
                self.visualizer.plot_heatmaps()
        else:
            print("\n---------skipped visualization-------------")
        
        print("\n---------RRAP complete-------------")


    def add_arguments(self):
        # TODO specify argument groups
        self.inputs = self.p.add_argument_group("## input arguments")
        self.outputs = self.p.add_argument_group("## output arguments")
        self.options = self.p.add_argument_group("## options")

        self.arg_groups.extend([self.inputs, self.outputs, self.optional])

        # Add the arguments
        self.inputs.add_argument('-i', help='text file of all dir paths that contain cleaned metaG fna files', 
                                 required=True)
        self.inputs.add_argument('-crg', help=' path for concatenated reference genome fa file', 
                                 required=False)
        self.inputs.add_argument('-rg', help='input directory for reference genomes', 
                                 required=True)
        self.outputs.add_argument('-o', help='output directory path', 
                                  required=True)
        self.inputs.add_argument('-n', help='name of the project', required=True, metavar='project name')

        self.inputs.add_argument('-sort_gen', help='txt file of sorted genomes (if --extra-vis flag is used)', required=False)
        self.inputs.add_argument('-sort_samples', help='txt file of sorted samples (if --extra-vis flag is used)', required=False)
        self.inputs.add_argument("--threads", help='number of available threads', required=False)
        self.inputs.add_argument("-suffix", default="_pass_1.fastq", 
                                  help="everything in metaG file name that is after the acc for the forward (R1) read files \n"
                                  "e.g. (-QUALITY_PASSED_R1.fastq for <sample_acc>-QUALITY_PASSED_R1.fastq) \n" 
                                  "Otherwise, RRAP assumes that the forward pass file name is formatted as <acc>_pass_1.fastq"
                                  "NOTE: suffixes that contain a dash must specify '--' as an escape character e.g. '-suffix \"-- -QUALITY_PASSED_R1.fastq\"'")

        # specify optional args
        self.options.add_argument("--skip-merge-contigs", default=True, dest='contig_merge',
                                  action='store_false', help="Concatenate contigs under individual organisms")
        self.options.add_argument("--skip-indexing", default=False, dest='index_pass',
                                  action='store_true', 
                                  help='Specify if the indexing step has already been completed and can be skipped. \
                                        If this flag is used, please check that bowtie2 index files exist e.g. \
                                        <output_dir_path>/index_dir/<project_name>.x.bt2')
        self.options.add_argument("--skip-rr", default=False, dest='rr_pass',
                                  action='store_true',
                                  help='Specify if the read recruitment step has already been completed and can be' \
                                       'skipped. If this flag is used, please check that read recruitment files exist' \
                                       'e.g. <output_dir_path>/')
        self.options.add_argument("--skip-vis", default=False, dest='vis_pass',
                                  action='store_true',
                                  help='Specify if the visualization step can be skipped')
        self.options.add_argument("--extra-vis", default=False, dest='extra_vis',
                                  action='store_true', 
                                  help='create csv with normalized RPKM values (log10) and plot heatmap using normalized values')
        self.options.add_argument("-q", default=True, dest='verbosity',
                                  action='store_false', help="more verbose output in terms of what the program is doing")

        self.args = self.p.parse_args()
        

    def set_output_dir(self):
        if self.args.o:
            # create output dir and create inner stats and index dir
            self.stats_dir_path = os.path.join(self.args.o, "stats", self.args.n)
            self.index_dir_path = os.path.join(self.args.o, "index", self.args.n)
            self.bam_dir_path = os.path.join(self.args.o, "bam", self.args.n)

            # make output (if it does not exist)
            if not os.path.isdir(self.args.o):
                subprocess.run("mkdir " + self.args.o, shell=True)

            # make stats, index, and bam dir
            if not os.path.isdir(os.path.join(self.args.o, "stats")):
                subprocess.run("mkdir " + os.path.join(self.args.o, "stats"), shell=True)
            if not os.path.isdir(os.path.join(self.args.o, "index")):
                subprocess.run("mkdir " + os.path.join(self.args.o, "index"), shell=True)
            if not os.path.isdir(os.path.join(self.args.o, "bam")):
                subprocess.run("mkdir " + os.path.join(self.args.o, "bam"), shell=True)

            # make project name specific dirs
            if not os.path.isdir(self.stats_dir_path):
                subprocess.run("mkdir " + self.stats_dir_path, shell=True)
            if not os.path.isdir(self.index_dir_path):
                subprocess.run("mkdir " + self.index_dir_path, shell=True)
            if not os.path.isdir(self.bam_dir_path):
                subprocess.run("mkdir " + self.bam_dir_path, shell=True)


if __name__ == "__main__":
    main()



import subprocess
import os
from pathlib import Path

class ReadRecruiter:
    def __init__(self, args, index_dir_path, cat_file_path, stats_dir_path, bam_dir_path):
        self.args = args
        self.index_dir_path = index_dir_path
        self.cat_file_path = cat_file_path
        self.stats_dir_path = stats_dir_path
        self.bam_dir_path = bam_dir_path

        # detect -- escape characters
        if self.args.suffix[0:2] == "--":
            self.args.suffix = self.args.suffix[3:]

    def read_recruit(self):
        if self.args.i:

            # retrieve individual clean dir paths from -i flag
            with open(self.args.i) as file:
                clean_dir_paths = file.readlines()
                clean_dir_paths = [line.rstrip() for line in clean_dir_paths]

            for clean_dir_path in clean_dir_paths:
                # create list of tuples with each tuple containing the forward and reverse file name for a sample
                tuple_list = self.find_acc(clean_dir_path)

                print("\nRunning read recruitment for dir: " + clean_dir_path)
                self.align_reads(tuple_list, clean_dir_path)

                print("\nread recruitment complete: " + clean_dir_path)
        else:
            pass

    def find_acc(self, path):
        # get list of all fastq files (r1 and r2)
        # self.args.suffix.split(".")[-1] denotes the fastq extension e.g.fastq or fq
        cmd = subprocess.run('ls {0}'.format(os.path.join(path, "*" + self.args.suffix.split(".")[-1])),
                             shell=True, capture_output=True, text=True)
        acc_list = cmd.stdout.split()
        acc_list = [line.rstrip() for line in acc_list]

        # make sure that half the fastq files contain the forward pass (r1) suffix

        contains_suffix = [sample for sample in acc_list if self.args.suffix in sample]
        if len(contains_suffix) == 0:
            raise IOError("Incorrect suffix ({0}) specified for forward pass fastq files ".format(self.args.suffix) +
                          "in path {0} Check the -suffix argument".format(path))
        elif len(contains_suffix) * 2 != len(acc_list):
            raise IOError("Incorrect suffix ({0}) specified for forward pass fastq files ".format(self.args.suffix) +
                          "in path {0}. Check that all fastq files have the same suffix.".format(self.args.path) +
                          "Check the -suffix argument")

        # sort list alphanumerically
        acc_list.sort()

        # create list of tuples with each tuple containing an even/odd pair e.g. (0, 1), (2, 3), (4, 5)
        tuple_list = []
        for i in range(int(len(acc_list)/2)):
            # general format of file name is <acc><suffix>
            acc = acc_list[i*2].split(self.args.suffix)[0]

            # acc_list[i*2] should denote r1 file, acc_list[i*2] should denote r2 file 
            tuple_list.append((acc, acc_list[i*2], acc_list[(i*2)+1]))

        return tuple_list


    def align_reads(self, tuple_list, clean_dir_path):
        # for each acc
        for sample in tuple_list:
            # create helpful variables
            acc = os.path.basename(sample[0])
            acc_bam_path_stem = os.path.join(self.bam_dir_path, acc)

            quiet_addon = ""
            if self.args.verbosity:
                print("\nworking on sample:", acc, "\n")
            else:
                quiet_addon = "--quiet"

            # only run bowtie2 if .bam.stats file doesn't exist
            if not os.path.exists(os.path.expanduser(os.path.join(self.stats_dir_path, "{0}.bam.stats".format(acc)))):
                # run read recruitment
                threads_addon = ""
                if self.args.threads:
                    threads_addon = "--threads {}".format(self.args.threads)

                command = 'bowtie2 {5} {0} -x "{1}" -1 "{2}" -2 "{3}" ' \
                          '--no-unal -S "{4}.sam"'.format(threads_addon, self.index_dir_path, sample[1], sample[2], 
                                                              acc_bam_path_stem, quiet_addon)
                if self.args.verbosity:
                    print("mapping reads with the following command: {}\n".format(command))
                subprocess.run(command, shell=True)

                # generate stats_dir
                self.generate_stats_file(acc_bam_path_stem)
            else:
                if self.args.verbosity:
                    print("previous file exists for", acc, "not running read recruitment")


    def generate_stats_file(self, acc):
        # convert output SAM fle into a BAM file
        subprocess.run('samtools view -F 4 -bS "{0}.sam" > "{0}-RAW.bam"'.format(acc), shell=True)

        # sort and index the BAM file
        subprocess.run('samtools sort "{0}-RAW.bam" -o "{0}.bam"'.format(acc), shell=True)
        subprocess.run('samtools index "{0}.bam"'.format(acc), shell=True)
        subprocess.run('samtools idxstats "{0}.bam" > "{0}.bam.stats"'.format(acc), shell=True)

        # copy bam.stats files into stats dir
        subprocess.run('mv {0}.bam.stats {1}'.format(acc, self.stats_dir_path), shell=True)

        # TODO remove temporary files
        subprocess.run('rm "{0}.sam" "{0}-RAW.bam"'.format(acc), shell=True)



import subprocess
import os
import pandas as pd

class Visualizer:
    def __init__(self, args, rpkm_heater_path, stats_dir_path):
        self.args = args
        self.rpkm_heater_path = rpkm_heater_path
        self.stats_dir_path = stats_dir_path

    def visualize(self):
        self.plot_heatmaps()

    def plot_heatmaps(self):
        if self.args.verbosity:
            print("calculating RPKM values with log normalization and generating heatmap with rpkm_heater:")

        # make appropriate dir
        rpkm_output_dir = self.create_rpkm_output_dir()

        sort_gen_addon = ""
        sort_samples_addon = ""

        # incorporate optional sort_gen and sort_sample options for rpkm_heater
        if self.args.sort_gen:
            sort_gen_addon = "-sort_gen {0}".format(self.args.sort_gen)
        if self.args.sort_samples:
            sort_samples_addon = "-sort_samples {0}".format(self.args.sort_samples)

        cmd = "rpkm_heater -map -i {1} -o {2} -project {3} {4} {5}".format(self.rpkm_heater_path,
                                                                           self.stats_dir_path,
                                                                           rpkm_output_dir,
                                                                           self.args.n,
                                                                           sort_gen_addon,
                                                                           sort_samples_addon)
        if self.args.verbosity:
            print("running: " + cmd)

        subprocess.run(cmd, shell=True)

    def calculate_rpkm(self):
        if self.args.verbosity:
            print("calculating RPKM values without log normalization:")
        # make appropriate dir
        rpkm_output_dir = self.create_rpkm_output_dir()

        # df holds rpkm values with genome acc as the row names and metaG acc as the headers
        df = pd.DataFrame()

        # loop through files with bam.stats suffix in generated stats dir
        for file in os.listdir(self.stats_dir_path):
            if file.endswith(".bam.stats"):

                # rpkm dict will hold rpkm values for single metaG
                rpkm = {}
                # read csv
                entry = pd.read_csv(os.path.join(self.stats_dir_path, file), delimiter="\t", header=None)
                # add headers to pd
                entry.columns = ["genome", "gen_length", "r_mapped", "r_unmapped"]
                # calculate total mapped reads
                tot_reads = entry['r_mapped'].sum()

                # specify metaG accession in dict
                rpkm['ACC'] = [file[:-10]]

                # specify rpkm values for each genome for this specific metaG
                for i in range(len(entry['genome'])):
                    if entry['genome'][i] != "*":
                        rpkm[entry['genome'][i]] = \
                            [entry['r_mapped'][i]/((entry['gen_length'][i]/1000)*(tot_reads/1000000))]

                # convert dict to data frame and transpose
                rpkm_df = pd.DataFrame(rpkm)
                rpkm_df.set_index('ACC', inplace=True)
                rpkm_df = rpkm_df.transpose()

                # add rpkm data to overall dataframe
                df = pd.concat([df, rpkm_df], axis=1, join='outer')

        if self.args.verbosity:
            print(df, "\n")
        df.to_csv(os.path.join(rpkm_output_dir, self.args.n + "_rpkm_noLog.csv"), index_label='ACC')

    def create_rpkm_output_dir(self):
        # make appropriate dir
        rpkm_base_dir = os.path.join(self.args.o, "rpkm")
        rpkm_output_dir = os.path.join(rpkm_base_dir, self.args.n)

        if not os.path.isdir(rpkm_base_dir):
            subprocess.run("mkdir {}".format(rpkm_base_dir), shell=True)
        if not os.path.isdir(rpkm_output_dir):  
            subprocess.run("mkdir {}".format(rpkm_output_dir), shell=True)
        
        return rpkm_output_dir




from setuptools import setup

setup(name='rrap',
      version='1.0.15',
      description='A metagenomic read recruitment data pipeline',
      url='https://github.com/Kojiconner/rrap_metag_pkg/',
      author='Conner Kojima',
      author_email='cykojima@usc.edu',
      license='MIT',
      packages=['rrap'],
      scripts=['bin/rrap', 'bin/rpkm_heater', 'bin/rrap_test'],
      install_requires=[
        "joblib==0.16.0",
        "scikit-learn==0.23.1",
        "scipy>=1.5.2",
        "threadpoolctl==2.1.0"
      ],
      python_requires=">=3.7.3",
      package_dir={"": "src"},
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
      ],
      zip_safe=False)



import subprocess
import os
from pathlib import Path

class Concatenator:
    def __init__(self, args):
        self.args = args
        self.cat_file_name = os.path.join(self.args.o, 'allgenomes_cat_{0}.fna'.format(self.args.n))

    def concatenate(self):
        # replace contigs with file name

        # delete cat file if it already exists
        if os.path.exists(self.cat_file_name):
            if self.args.verbosity:
                print("deleting preexisting concatenated file: {0}".format(self.cat_file_name))
            subprocess.run('rm -f {0}'.format(self.cat_file_name), shell=True)

        if self.args.verbosity:
            print("concatenating genomes from the dir {0}".format(self.args.rg))

        if self.args.contig_merge:
            # get list of all ref genome file paths
            cmd = subprocess.run('ls {0}'.format(os.path.join(self.args.rg, "*")),
                                    shell=True, capture_output=True, text=True)
            rf_list = cmd.stdout.split()

            # read into each rf genome and change headers to group contigs under organism header (file name)
            for rf_path in rf_list:
                self.combine_contigs(rf_path)
        else:
            # concatenate all ref genomes into one file
            subprocess.run("cat {0} > {1}".format(os.path.join(self.args.rg, "*"), self.cat_file_name), shell=True)
        return self.cat_file_name

    def combine_contigs(self, rf_path):
        # replace initial header line with line that contains just organism e.g. (>Candidatus Fonsibacter)
        # remove all subsequent header lines
        # this line remains in place to deal with files that contain many contigs
        rf_name = Path(rf_path).stem

        # read rf genome and add initial header/remove contig headers
        with open(rf_path) as file:
            lines = file.readlines()
            lines = [line for line in lines[1:] if not line.startswith(">")]
            lines.insert(0, ">" + rf_name + "\n")

        # write modified lines into concatenated fle
        with open(self.cat_file_name, 'a') as f:
            for line in lines:
                f.write("%s" % line)
